# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RabbitMQ Eiffel publisher."""
import time
import logging
import pika
from eiffellib.publishers.eiffel_publisher import EiffelPublisher
from eiffellib.lib.base_rabbitmq import BaseRabbitMQ

_LOG = logging.getLogger(__name__)


class RabbitMQPublisher(EiffelPublisher, BaseRabbitMQ):
    """Rabbitmq connection for sending messages."""
    _acks = 0
    _nacks = 0
    _delivered = 0
    _last_delivered_tag = 0

    # pylint:disable=too-many-arguments
    def __init__(self, host, exchange, routing_key="eiffel",
                 username=None, password=None, port=5671, vhost=None,
                 source=None, ssl=True):
        """Initialize with host and create pika connection parameters."""
        BaseRabbitMQ.__init__(self, host, port, username, password, vhost, ssl)
        self._deliveries = {}
        self._nacked_deliveries = []
        self.exchange = exchange
        self.routing_key = routing_key
        self.source = source

    # Tell EiffelPublisher to use BaseRabbitMQ.start
    start = BaseRabbitMQ.start

    def reset_parameters(self):
        """Reset parameters to default."""
        super().reset_parameters()
        self._last_delivered_tag = 0
        self._delivered = 0

    def _setup(self, channel):
        """Start the RabbitMQ publisher.

        :param channel: RabbitMQ channel.
        :type channel: :obj:`pika.channel.Channel`
        """
        self._channel.confirm_delivery(self._confirm_delivery)
        self._start()

    def _start(self, *args, **kwargs):
        """Start the RabbitMQ publisher."""
        _LOG.info("Start publishing messages.")
        self._channel.add_on_cancel_callback(self._publisher_canceled)

        # If the server shut down due to broker failure, attempt to recover lost messages.
        self._nacked_deliveries.extend(self._deliveries.values())
        self._deliveries.clear()
        self._connection.ioloop.call_later(1, self._resend_nacked_deliveries)

        self._was_active = True
        self.active = True
        self.running = True

    def _cancel(self):
        """Close channel and set active to False."""
        if self._channel:
            self.active = False
            self.close_channel()

    def _publisher_canceled(self, method_frame):
        """Publisher remotely canceled callback.

        :param method_frame: Frame which triggered this callback.
        :type method_frame: class
        """
        _LOG.info("Publisher was canceled remotely, shutting down: %r", method_frame)
        if self._channel:
            self._channel.close()

    def _resend_nacked_deliveries(self):
        """Resend all NACKed deliveries. This method loops forever."""
        deliveries = self._nacked_deliveries.copy()
        if deliveries:
            _LOG.info("Resending %i NACKed deliveries", len(deliveries))
        self._nacked_deliveries.clear()
        for event in deliveries:
            self.send_event(event)
        self._connection.ioloop.call_later(1, self._resend_nacked_deliveries)

    def _confirm_delivery(self, method_frame):
        """Confirm the delivery of events and make sure we resend NACKed events.

        Note: This only checks for NACKs from the broker, not from any consumer.

        :param method_frame: Frame which triggered this callback.
        :type method_frame: class
        """
        method = method_frame.method
        confirmation_type = method.NAME.split('.')[1].lower()
        delivery_tag = method.delivery_tag
        multiple = method.multiple

        if multiple and delivery_tag == 0:
            number_of_acks = len(self._deliveries)
        else:
            number_of_acks = delivery_tag - self._last_delivered_tag

        if confirmation_type == 'ack':
            self._acks += number_of_acks
        elif confirmation_type == 'nack':
            self._nacks += number_of_acks

        if delivery_tag == 0:
            if confirmation_type == "nack":
                self._nacked_deliveries.extend(self._deliveries.values())
            self._deliveries.clear()
        else:
            for tag in range(self._last_delivered_tag + 1, delivery_tag + 1):
                if confirmation_type == "nack":
                    self._nacked_deliveries.append(self._deliveries[tag])
                self._deliveries.pop(tag)
        self._last_delivered_tag = delivery_tag

        _LOG.debug('Published %i messages, %i have yet to be confirmed, '
                   '%i were acked and %i were nacked', self._acks+self._nacks,
                   len(self._deliveries), self._acks, self._nacks)

    def send_event(self, event, block=True):
        """Validate and send an eiffel event to the rabbitmq server.

        :param event: Event to send.
        :type event: :obj:`eiffellib.events.eiffel_base_event.EiffelBaseEvent`
        :param block: Set to True in order to block for channel to become ready.
                      Default: True
        :type block: bool
        """
        if block:
            self.wait_start()
            while self._channel is None or not self._channel.is_open:
                time.sleep(0.1)

        correlation_id = event.meta.event_id
        properties = pika.BasicProperties(content_type="application/json",
                                          delivery_mode=2)
        if self.source is not None:
            event.meta.add("source", self.source)
        event.validate()
        try:
            self._channel.basic_publish(
                self.exchange,
                self.routing_key,
                event.serialized,
                properties,
            )
        except:
            self._nacked_deliveries.append(event)
            return
        self._delivered += 1
        self._deliveries[self._delivered] = event

    send = send_event



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Eiffel base publisher."""


class EiffelPublisher():
    """Base connection class for sending messages."""

    connection = None
    started = False
    running = False

    def start(self):
        """Start the pika rabbitmq connection."""
        raise NotImplementedError

    def send_event(self, event, block=True):
        """Validate and send an eiffel event to server.

        :param block: Set to True in order to block until ready.
                      Default: True
        :type block: bool
        :param event: Event to send.
        :type event: :obj:`eiffellib.events.eiffel_base_event.EiffelBaseEvent`
        """
        raise NotImplementedError

    def send(self, msg):
        """Send a message to the server.

        :param msg: Message to send.
        :type msg: str
        """
        raise NotImplementedError

    def close(self):
        """Close down publisher. Override if special actions are required."""
        self.running = False



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RabbitMQ Eiffel subscriber."""
import logging
import threading
from multiprocessing.pool import ThreadPool
import functools
import pika
from eiffellib.subscribers.eiffel_subscriber import EiffelSubscriber
from eiffellib.lib.base_rabbitmq import BaseRabbitMQ

_LOG = logging.getLogger(__name__)


# pylint:disable=too-many-instance-attributes
class RabbitMQSubscriber(EiffelSubscriber, BaseRabbitMQ):
    """Receiver for Rabbit MQ event databases."""

    prefetch_count = 4   # RabbitMQ QOS prefetch count.
    max_threads = 100    # Max number of callback threads active.
    max_queue = 100      # Max number of waiting callbacks.

    _consumer_tag = None

    __thread_pool = None
    __workers = None

    # pylint:disable=too-many-arguments
    def __init__(self, host, queue, exchange, username=None, password=None, port=5671,
                 vhost=None, routing_key=None, ssl=True, queue_params=None):
        """Initialize with rabbitmq host."""
        super(RabbitMQSubscriber, self).__init__()
        BaseRabbitMQ.__init__(self, host, port, username, password, vhost, ssl)
        self.host = host
        self.queue = queue
        self.exchange = exchange
        self.routing_key = routing_key or "#"
        self.queue_params = queue_params if queue_params else {}

    def reset_parameters(self):
        """Reset parameters to default."""
        super().reset_parameters()
        self._consumer_tag = None

    def _setup(self, channel):
        """Set up channel. Declare a queue to listen to.

        :param channel: RabbitMQ channel.
        :type channel: :obj:`pika.channel.Channel`
        """
        _LOG.info("Declaring queue %r", self.queue)
        callback = functools.partial(self._queue_declared, queue_name=self.queue)
        channel.queue_declare(queue=self.queue, callback=callback, **self.queue_params)

    def _start(self, *args, **kwargs):
        """Start consuming messages."""
        _LOG.info("QOS set to: %d", self.prefetch_count)
        _LOG.info("Start consuming messages.")
        self._channel.add_on_cancel_callback(self._consumer_canceled)

        self._consumer_tag = self._channel.basic_consume(self.queue, self._on_message)
        self._was_active = True
        self.active = True
        self.running = True

    def _cancel(self):
        """Send a Cancel request to RabbitMQ."""
        if self._channel:
            _LOG.info("Sending Basic.Cancel to RabbitMQ")
            callback = functools.partial(self._cancel_consumer, consumer_tag=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, callback)

    def _on_start(self):
        """Setup ThreadPool and Semaphore lock before starting."""
        self.__thread_pool = ThreadPool(self.max_threads)
        self.__workers = threading.Semaphore(self.max_threads + self.max_queue)

    def _queue_declared(self, _, queue_name):
        """Queue declared callback. Bind queue.

        :param queue_name: Name of the queue that was declared.
        :type queue_name: str
        """
        _LOG.info("Binding %r to %r with routing key %r", self.exchange,
                  queue_name, self.routing_key)
        callback = functools.partial(self._queue_bound, queue_name=queue_name)
        if self.routing_key is not None:
            self._channel.queue_bind(exchange=self.exchange, queue=queue_name,
                                     routing_key=self.routing_key, callback=callback)
        else:
            self._channel.queue_bind(exchange=self.exchange, queue=self.queue, callback=callback)

    def _queue_bound(self, _, queue_name):
        """Queue bound callback. Set QOS."""
        _LOG.info("Queue bound: %r", queue_name)
        self._channel.basic_qos(prefetch_count=self.prefetch_count, callback=self._start)

    def _consumer_canceled(self, method_frame):
        """Channel remotely canceled callback.

        :param method_frame: Frame which triggered this callback.
        :type method_frame: class
        """
        _LOG.info("Consumer was canceled remotely, shutting down: %r", method_frame)
        if self._channel:
            self._channel.close()

    def _on_message(self, _, method, __, body):
        """On message callback. Called on each message. Will block if no place in queue.

        For each message attempt to acquire the `threading.Semaphore`. The semaphore
        size is `max_threads` + `max_queue`. This is to limit the amount of threads
        in the queue, waiting to be processed.
        For each message apply them async to a `ThreadPool` with size=`max_threads`.

        :param method: Pika basic deliver object.
        :type method: :obj:`pika.spec.Basic.Deliver`
        :param properties: Pika basic properties object.
        :type properties: :obj:`pika.spec.BasicProperties`
        :param body: Message body.
        :type body: bytes
        """
        self.__workers.acquire()
        delivery_tag = method.delivery_tag
        error_callback = functools.partial(self.callback_error, delivery_tag)
        result_callback = functools.partial(self.callback_results, delivery_tag)
        self.__thread_pool.apply_async(self.call, args=(body,), callback=result_callback,
                                       error_callback=error_callback)

    def callback_results(self, delivery_tag, result):
        """Result callback for the ThreadPool. Called on successful execution of event.

        Add a callback to the ioloop depending on the result of execution.
        If result is ack, call :meth:`acknowledge`.
        If result is requeue, call :meth:`requeue`.
        If result is not ack and not requeue, call :meth:`reject`.

        :param delivery_tag: Delivery tag for the message that triggered.
        :type delivery_tag: int
        :param result: Result object for ThreadPool.
        :type result: :obj:`multiprocessing.pool.AsyncResult`
        """
        self.__workers.release()
        ack, requeue = result
        if ack:
            callback = functools.partial(self.acknowledge, self._channel, delivery_tag)
        elif requeue:
            callback = functools.partial(self.requeue, self._channel, delivery_tag)
        else:
            callback = functools.partial(self.reject, self._channel, delivery_tag)
        self._connection.ioloop.add_callback(callback)

    def callback_error(self, delivery_tag, exception):
        """Error callback for the ThreadPool. Reject the message.

        :param delivery_tag: Delivery tag for the message that triggered.
        :type delivery_tag: int
        :param exception: Exception raised from within event callback.
        :type exception: Exception
        """
        _LOG.warning("Callback raised exception: %r", exception)
        self.__workers.release()
        callback = functools.partial(self.reject, self._channel, delivery_tag)
        self._connection.ioloop.add_callback(callback)

    @staticmethod
    def acknowledge(channel, delivery_tag):
        """Acknowledge that a message has been received and will be processed.

        :param channel: Channel to acknowledge on.
        :type channel: :obj:`pika.channel.Channel`
        :param delivery_tag: Delivery tag to acknowledge.
        :type delivery_tag: int
        """
        try:
            channel.basic_ack(delivery_tag)
        except pika.exceptions.AMQPChannelError as exception:
            _LOG.error("Exception when attempting to ACK: %r", exception)

    @staticmethod
    def reject(channel, delivery_tag):
        """Reject a message and remove it completely from RabbitMQ.

        :param channel: Channel to reject on.
        :type channel: :obj:`pika.channel.Channel`
        :param delivery_tag: Delivery tag to reject.
        :type delivery_tag: int
        """
        try:
            channel.basic_reject(delivery_tag, requeue=False)
        except pika.exceptions.AMQPChannelError as exception:
            _LOG.error("Exception when attempting to REJECT: %r", exception)

    @staticmethod
    def requeue(channel, delivery_tag):
        """Requeue a message based on delivery_tag.

        :param channel: Channel to requeue on.
        :type channel: :obj:`pika.channel.Channel`
        :param delivery_tag: Delivery tag to requeue.
        :type delivery_tag: int
        """
        try:
            channel.basic_reject(delivery_tag, requeue=True)
        except pika.exceptions.AMQPChannelError as exception:
            _LOG.error("Exception when attempting to REQUEUE: %r", exception)

    def _cancel_consumer(self, _, consumer_tag):
        """Consumer canceled callback. RabbitMQ acknowledged the cancelation. Close channel.

        :param consumer_tag: Consumer tag for this subscriber.
        :type consumer_tag: str
        """
        self.active = False
        _LOG.info("RabbitMQ acknowledged the cancelation of the consumer: %r", consumer_tag)
        self.close_channel()



# Copyright 2019 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Activity module."""
from eiffellib.events import (EiffelActivityTriggeredEvent, EiffelActivityStartedEvent,
                              EiffelActivityFinishedEvent, EiffelActivityCanceledEvent)


# pylint: disable=too-many-instance-attributes
class Activity():
    """A base activity callable that can be used as a callback when subscribing to an event.

    This class provides an interface where it sends all the activity events that
    are necessary.
    """

    triggered = None
    started = None
    finished = None
    canceled = None

    def __init__(self, activity_name, publisher, source, triggers=None, execution_type=None):  # pylint:disable=too-many-arguments
        """Initialize.

        :param activity_name: Name of the activity in question.
        :type activity_name: str
        :param publisher: The publisher to use for sending events.
        :type publisher: :obj:`eiffellib.publisher.eiffel_publisher.EiffelPublisher`
        :param source: Source of the activity.
                       https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelActivityTriggeredEvent.md#metasource
        :type source: dict
        :param triggers: A list of dictionaries describing the triggers for the activity.
                         Note: Dictionary must contain the "type" key.
        :type triggers: list
        :param execution_type: Which type of execution.
                               Example: MANUAL, SEMI_AUTOMATED, AUTOMATED, OTHER
        :type execution_type: str
        """
        self.activity_name = activity_name
        self.triggers = triggers
        self.execution_type = execution_type
        self.publisher = publisher
        self.source = source

    def activity_triggered(self, context):
        """Send an activity triggered.

        :param context: Context for the event received.
        :type context: str
        """
        self.triggered = EiffelActivityTriggeredEvent()
        self.triggered.data.add("name", self.activity_name)
        if self.triggers is not None:
            self.triggered.data.add("triggers", self.triggers)
        if self.execution_type is not None:
            self.triggered.data.add("executionType", self.execution_type)
        if context is not None:
            self.triggered.links.add("CONTEXT", context)
        self.triggered.meta.add("source", self.source)
        self.publisher.send_event(self.triggered)

    def activity_started(self, context):
        """Send an activity started. Canceled if activity was not triggered.

        :param context: Context for the event received.
        :type context: str
        """
        if self.triggered is None:
            self.activity_canceled(context, "Activity not triggered.")
            return

        self.started = EiffelActivityStartedEvent()
        if context is not None:
            self.started.links.add("CONTEXT", context)
        self.started.links.add("ACTIVITY_EXECUTION", self.triggered)
        self.started.meta.add("source", self.source)
        self.publisher.send_event(self.started)

    def activity_finished(self, context):
        """Send an activity finished. Canceled if activity was not triggered.

        :param context: Context for the event received.
        :type context: str
        """
        if self.started is None:
            self.activity_canceled(context, "Activity not started")
            return

        self.finished = EiffelActivityFinishedEvent()
        self.finished.data.add("outcome", {"conclusion": "SUCCESSFUL"})
        if context is not None:
            self.finished.links.add("CONTEXT", context)
        self.finished.links.add("ACTIVITY_EXECUTION", self.triggered)
        self.finished.meta.add("source", self.source)
        self.publisher.send_event(self.finished)

    def activity_canceled(self, context, reason=None):
        """Send an activity canceled.

        :param context: Context for the event received.
        :type context: str
        :param reason: Reason for cancelation. Optional.
        :type reason: str
        """
        self.canceled = EiffelActivityCanceledEvent()
        if reason is not None:
            self.canceled.data.add("reason", reason)
        if context is not None:
            self.canceled.links.add("CONTEXT", context)
        self.canceled.links.add("ACTIVITY_EXECUTION", self.triggered)
        self.canceled.meta.add("source", self.source)
        self.publisher.send_event(self.canceled)

    def __call__(self, event, context):
        """Use as callback method for events.

        :param event: Event for the callback.
        :type event: :obj:`eiffellib.events.base_event.BaseEvent`
        """
        try:
            self.activity_triggered(context)
            call_context = context if context else self.triggered.meta.event_id
            self.pre_call(event, call_context)
            self.activity_started(context)
            self.call(event, call_context)
            self.post_call(event, call_context)
            self.activity_finished(context)
        except Exception as exception:
            self.activity_canceled(context, str(exception))
            raise

    def pre_call(self, event, context):
        """Execute before the 'call' method and just after activity triggered.

        :param event: Event for the callback.
        :type event: :obj:`eiffellib.events.base_event.BaseEvent`
        :param context: Context for the event received.
        :type context: str
        """

    def call(self, event, context):
        """Execute after the 'pre_call' method and just after activity started.

        :param event: Event for the callback.
        :type event: :obj:`eiffellib.events.base_event.BaseEvent`
        :param context: Context for the event received.
        :type context: str
        """

    def post_call(self, event, context):
        """Execute after the 'call' method and just before activity finished.

        :param event: Event for the callback.
        :type event: :obj:`eiffellib.events.base_event.BaseEvent`
        :param context: Context for the event received.
        :type context: str
        """



# Copyright 2019 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=line-too-long
from .eiffel_activity_canceled_event import EiffelActivityCanceledEvent
from .eiffel_activity_finished_event import EiffelActivityFinishedEvent
from .eiffel_activity_started_event import EiffelActivityStartedEvent
from .eiffel_activity_triggered_event import EiffelActivityTriggeredEvent
from .eiffel_announcement_published_event import EiffelAnnouncementPublishedEvent
from .eiffel_artifact_created_event import EiffelArtifactCreatedEvent
from .eiffel_artifact_published_event import EiffelArtifactPublishedEvent
from .eiffel_artifact_reused_event import EiffelArtifactReusedEvent
from .eiffel_composition_defined_event import EiffelCompositionDefinedEvent
from .eiffel_confidence_level_modified_event import EiffelConfidenceLevelModifiedEvent
from .eiffel_environment_defined_event import EiffelEnvironmentDefinedEvent
from .eiffel_flow_context_defined_event import EiffelFlowContextDefinedEvent
from .eiffel_issue_defined_event import EiffelIssueDefinedEvent
from .eiffel_issue_verified_event import EiffelIssueVerifiedEvent
from .eiffel_source_change_created_event import EiffelSourceChangeCreatedEvent
from .eiffel_source_change_submitted_event import EiffelSourceChangeSubmittedEvent
from .eiffel_test_case_canceled_event import EiffelTestCaseCanceledEvent
from .eiffel_test_case_finished_event import EiffelTestCaseFinishedEvent
from .eiffel_test_case_started_event import EiffelTestCaseStartedEvent
from .eiffel_test_case_triggered_event import EiffelTestCaseTriggeredEvent
from .eiffel_test_execution_recipe_collection_created_event import EiffelTestExecutionRecipeCollectionCreatedEvent  # noqa
from .eiffel_test_suite_started_event import EiffelTestSuiteStartedEvent
from .eiffel_test_suite_finished_event import EiffelTestSuiteFinishedEvent



# Copyright 2019 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base Eiffel event."""
import os
import json
import uuid
import time
from jsonschema import validate
from eiffellib import BASE_PATH

# We're inheriting 'object' which is unnecessary with python3,
# however since this package shall be usable with both python2
# and python3 we still need to keep these.
# pylint:disable=useless-object-inheritance


class EiffelBaseMeta(object):
    """Eiffel base meta object."""

    def __init__(self, _type, version):
        """Initialize with event type and version."""
        self.type = _type
        self.version = version
        self.event_id = str(uuid.uuid4())
        self.time = int(round(time.time() * 1000))
        self.optional = []

    def add(self, key, value):
        """Add an optional meta parameter.

        :param key: Key to add.
        :type key: str
        :param value: Value for specific key.
        :type value: Any
        """
        self.optional.append((key, value))

    def rebuild(self, meta):
        """Rebuild meta object with new data.

        This can be used to rebuild an event using json data
        received from the eiffel consumer.

        :param meta: Meta data to rebuild with.
        :type meta: dict
        """
        self.type = meta.pop("type")
        self.version = meta.pop("version")
        self.time = meta.pop("time")
        self.event_id = meta.pop("id")
        for key, value in meta.items():
            self.add(key, value)

    @property
    def json(self):
        """Meta object as a dict.

        :return: This meta class as a json serializable dict.
        :rtype: dict
        """
        meta_json = {"type": self.type,
                     "version": self.version,
                     "id": str(self.event_id),
                     "time": self.time}
        for key, value in self.optional:
            meta_json[key] = value
        return meta_json


class EiffelBaseLink(object):
    """Eiffel base link object."""

    def __init__(self):
        """Initialize with a possible types dict and links list."""
        self.links = []

    def add(self, _type, target):
        """Validate and add link to class.

        Note: Only validates event type if target is an event.

        :param _type: Link type to add.
        :type _type: str
        :param target: Target for the link.
        :type target: :obj:`EiffelBaseEvent` or str
        """
        if isinstance(target, str):
            target_id = target
        else:
            target_id = target.meta.event_id
        self.links.append({"type": _type, "target": target_id})

    def rebuild(self, links):
        """Rebuild links object with new data.

        This can be used to rebuild an event using json data
        received from the eiffel consumer.

        :param links: Links data to rebuild with.
        :type links: list
        """
        self.links = []
        for link in links:
            self.add(link.get("type"), link.get("target"))

    @property
    def json(self):
        """Json serializable list of links.

        :return: List of links.
        :rtype: list
        """
        return self.links


class EiffelBaseData(object):
    """Eiffel base data object."""

    def __init__(self):
        """Initialize with a data dictionary."""
        self.data = {}

    def add(self, key, value):
        """Add data to object.

        :param key: Key to add data to.
        :type key: str
        :param value: Value for key.
        :type value: Any
        """
        self.data[key] = value

    def rebuild(self, data):
        """Rebuild data object with new data.

        This can be used to rebuild an event using json data
        received from the eiffel consumer.

        :param data: Data to rebuild with.
        :type data: dict
        """
        self.data = data

    @property
    def json(self):
        """Json serializable list of data.

        :return: Dictionary of data.
        :rtype: dict
        """
        return self.data


class EiffelBaseEvent(object):
    """Eiffel base event object."""

    schema_file = None
    __schema = None
    version = "0.0.1"
    meta = EiffelBaseMeta("EiffelBaseEvent", version)
    links = EiffelBaseLink()

    def __init__(self, version=None):
        """Initialize with a base data object."""
        if version is not None:
            self.version = version
        self.data = EiffelBaseData()
        self.load_schema(self.version)

    def rebuild(self, json_data):
        """Rebuild event using json data.

        Calls the meta, data and links objects rebuild methods
        with data input.
        This can be used to rebuild an event using json data
        received from the eiffel consumer.

        :param json_data: Data to rebuild with.
        :type json_data: dict
        """
        self.meta.rebuild(json_data.get("meta", {}))
        self.data.rebuild(json_data.get("data", {}))
        self.links.rebuild(json_data.get("links", []))
        self.load_schema(self.meta.version)
        self.validate()

    @property
    def json(self):
        """Json serializable eiffel event."""
        return {"meta": self.meta.json,
                "data": self.data.json,
                "links": self.links.json}

    def load_schema(self, version):
        """Load a schema file path based on event name and version.

        :param version: Version of eiffel event to validate against.
        :type version: str
        """
        self.schema_file = os.path.join(BASE_PATH, "schemas",
                                        self.__class__.__name__,
                                        "{}.json".format(version))

    @property
    def schema(self):
        """Json schema for the current event."""
        if not self.__schema:
            with open(self.schema_file) as schema_file:
                self.__schema = json.load(schema_file)
        return self.__schema

    def validate(self):
        """Validate the json data in the eiffel event.

        :raises: ValidationError.
        """
        validate(self.json, self.schema)

    @property
    def serialized(self):
        """Json data serialized to string.

        :return: Json string.
        :rtype: str
        """
        return json.dumps(self.json)

    @property
    def pretty(self):
        """Pretty version of the json data.

        :return: Pretty formatted json.
        :rtype: str
        """
        return json.dumps(self.json, indent=4, sort_keys=True)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RabbitMQ base connection."""
import time
import logging
import threading
import ssl as _ssl
import pika


_LOG = logging.getLogger(__name__)


# pylint:disable=too-many-instance-attributes
class BaseRabbitMQ:
    """Base RabbitMQ connection object."""

    _connection = None
    _channel = None
    _was_active = False
    _closing = False

    should_reconnect = False
    active = False
    connection_thread = None
    running = False

    # pylint:disable=too-many-arguments
    def __init__(self, host, port, username, password, vhost, ssl):
        """Set connection parameters."""
        parameters = {"port": port}
        if ssl is True:
            context = _ssl.create_default_context()
            ssl_options = pika.SSLOptions(context, host)
            parameters["ssl_options"] = ssl_options
        if username and password:
            parameters["credentials"] = pika.PlainCredentials(username, password)
        if vhost:
            parameters["virtual_host"] = vhost

        self.parameters = pika.ConnectionParameters(host, **parameters)

    def reset_parameters(self):
        """Reset parameters to default."""
        self.should_reconnect = False
        self.active = False
        self._closing = False
        self._was_active = False

    def _setup(self, channel):
        """Setup channel. Called after channel is opened.

        :param channel: RabbitMQ channel.
        :type channel: :obj:`pika.channel.Channel`
        """
        raise NotImplementedError

    def _start(self, *args, **kwargs):
        """Start up the RabbitMQ connection. Call when connection is ready."""
        raise NotImplementedError

    def _cancel(self):
        """Cancel the RabbitMQ connection."""
        raise NotImplementedError

    def _on_start(self):
        """Called just before starting the connection."""

    def _on_connect(self):
        """Called just before connecting."""

    def reconnect(self):
        """Reconnect RabbitMQ connection."""
        self.should_reconnect = True
        self.stop()

    def connect(self):
        """Connect to RabbitMQ."""
        self._connection = pika.SelectConnection(self.parameters,
                                                 on_open_callback=self._connection_open,
                                                 on_close_callback=self._connection_close,
                                                 on_open_error_callback=self._connection_open_error)

    def close_connection(self):
        """Close the RabbitMQ connection."""
        self.active = False
        if self._connection.is_closing or self._connection.is_closed:
            _LOG.info("Connection is closing or already closed.")
        else:
            _LOG.info("Closing connection")
            self._connection.close()

    def _connection_open(self, connection):
        """Connection opened callback. Create a RabbitMQ channel.

        :param connection: Connection that was just opened.
        :type connection: :obj:`pika.connection.Connection`
        """
        _LOG.info("Connection opened")
        _LOG.info("Creating a new channel.")
        connection.channel(on_open_callback=self._channel_open)

    def _connection_close(self, _, reason):
        """Connection closed callback.

        Close the connection if told to, otherwise attempt reconnect.

        :param connection: Connection that was closed.
        :type connection: :obj:`pika.connection.Connection`
        :param reason: Reason the connection was closed.
        :type reason: str
        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            _LOG.warning("Connection closed, reconnecting: %r", reason)
            self.reconnect()

    def _connection_open_error(self, _, error):
        """Connection error when starting callback. Reconnect.

        :param connection: Connection that was closed.
        :type connection: :obj:`pika.Connection`
        :param error: Reason the connection was closed.
        :type error: str
        """
        _LOG.error("Failed to open connection: %r", error)
        self.reconnect()

    def _channel_open(self, channel):
        """Channel opened callback. Set close callback and call setup channel.

        :param channel: Channel that was just opened.
        :type channel: :obj:`pika.channel.Channel`
        """
        _LOG.info("Channel opened")
        self._channel = channel
        self._channel.add_on_close_callback(self._channel_closed)
        self._setup(channel)

    def _channel_closed(self, channel, reason):
        """Channel closed callback. Close connection."""
        _LOG.warning("Channel %i was closed: %r", channel, reason)
        self.close_connection()

    def close_channel(self):
        """Close the RabbitMQ channel."""
        _LOG.info("Closing the channel.")
        self._channel.close()

    def run(self):
        """Reset parameters, connect to RabbitMQ and start the ioloop.

        This is a blocking method.
        """
        self._on_connect()
        self.reset_parameters()
        self.connect()
        self._connection.ioloop.start()

    def keep_alive(self):
        """Reconnect forever if the should_reconnect flag is set.

        This is a blocking method.
        """
        reconnect_delay = 0
        while True:
            try:
                self.run()
            except KeyboardInterrupt:
                self.stop()
                break
            if self.should_reconnect:
                self.stop()
                if self._was_active:
                    reconnect_delay = 0
                else:
                    reconnect_delay += 1
                if reconnect_delay > 30:
                    reconnect_delay = 30
                _LOG.info("Reconnecting after %d seconds", reconnect_delay)
                time.sleep(reconnect_delay)
            else:
                break
        self.running = False

    def is_alive(self):
        """Check if connection is alive or not.

        Note that this only checks the 'running' flag and does not
        deep-dive into the RabbitMQ connection and channel.

        :return: Whether this connection is running or not.
        :rtype: bool
        """
        return self.running

    def wait_start(self):
        """Block until connection starts."""
        while not self.is_alive():
            time.sleep(0.1)

    def wait_close(self):
        """Block until publisher closes."""
        while self.is_alive():
            time.sleep(0.1)

    def start(self, wait=True):
        """Start the RabbitMQ connection in a thread.

        :param wait: Block until the connection starts. Defaults to True.
        :type wait: bool
        """
        self._on_start()
        self.connection_thread = threading.Thread(target=self.keep_alive)
        self.connection_thread.daemon = True
        self.connection_thread.start()
        if wait:
            self.wait_start()

    def stop(self):
        """Stop the RabbitMQ connection."""
        if not self._closing:
            self._closing = True
            _LOG.info("Stopping")
            if self.active:
                self._cancel()
                try:
                    # Start the ioloop again in order for the callbacks
                    # to fire, so that we can do a clean shutdown.
                    self._connection.ioloop.start()
                except RuntimeError:
                    pass
            else:
                self._connection.ioloop.stop()
            _LOG.info("Stopped")

    close = stop  # Backwards compatibility
